@@ -12,10 +12,28 @@
 
 import uuid
 
+from fom.db import ITERABLE_TYPES, SERIALIZABLE_TYPES
 from fom.session import Fluid
 from fom.errors import Fluid404Error
 
 
+def _get_tag_values(cls):
+    """Given a class will return a list containing the name and instance of
+    all the attributes that are of type readonly_tag_value
+    """
+    # Step 1. Create the list of matching fields associated with this klass
+    tags = [(name, instance)
+        for (name, instance) in cls.__dict__.items()
+        if isinstance(instance, readonly_tag_value)]
+    # Step 2. recursively loop over the parent classes of klass so we don't
+    # miss any fields *they* might have defined (like "about" in the Object
+    # class)
+    for b in cls.__bases__:
+        if issubclass(b, Object):
+            tags.extend(_get_tag_values(b))
+    return tags
+
+
 class SessionBound(object):
     """Something with a path that is bound to a database.
 
@@ -34,10 +52,15 @@ class SessionBound(object):
             fluid = Fluid.bound
         self.fluid = fluid
 
+    def __repr__(self):
+        return '<%s path=%r>' % (self.__class__.__name__, self.path)
+
+
 class Policy(object):
     """
     Represents a policy specification for permissions on an action
     """
+
     def __init__(self, policy='open', exceptions=[]):
         self.policy = policy
         self.exceptions = exceptions
@@ -47,10 +70,12 @@ class Policy(object):
 
     __str__ = __repr__
 
+
 class Permissions(object):
     """
     A nicer view onto Namespace/Tag/TagValue permissions
     """
+
     def __init__(self, api):
         """ api references an appropriately created ItemPermissionsApi
         instance
@@ -71,20 +96,19 @@ class Permissions(object):
         else:
             raise TypeError('Use an instance of Policy to set the policy')
 
+
 class Namespace(SessionBound):
-    """A FluidDB Namespace"""
+    """A FluidDB Namespace
+    """
 
     def __init__(self, path, fluid=None):
         super(Namespace, self).__init__(path, fluid)
-        self.permissions = Permissions(self.fluid.permissions.namespaces[self.path])
+        self.permissions = Permissions(
+            self.fluid.permissions.namespaces[self.path])
 
     def create(self, description):
         """Create this namespace.
 
-        >>> ns = Namespace(u'test/reviews')
-        >>> ns.create(u'A place for all my reviews')
-        <FluidResponse (201, 'application/json' ...>
-
         :param description: The description of the Namespace
         """
         parent, name = path_split(self.path)
@@ -94,11 +118,6 @@ class Namespace(SessionBound):
     def create_namespace(self, name, description):
         """Create a child namespace, and return it.
 
-        >>> ns = Namespace(u'test')
-        >>> reviews_ns = ns.create_namespace(
-        ...     u'reviews', u'A place for all my reviews')
-
-
         :param name: The name of the Namespace to be created
         :param description: The description of the Namespace to be created
         """
@@ -106,10 +125,7 @@ class Namespace(SessionBound):
         return Namespace(path_child(self.path, name))
 
     def create_tag(self, name, description, indexed):
-        """Create a tag in this namespace.
-
-        >>> ns = Namespace(u'test')
-        >>> tag = ns.create_tag(u'review', u'A review I write', False)
+        """Create a tag in this namespace, and return it.
 
         :param name: The name of the Tag to be created
         :param description: The description of the Tag to be created
@@ -119,33 +135,21 @@ class Namespace(SessionBound):
         return Tag(path_child(self.path, name))
 
     def delete(self):
-        """Delete this namespace
-
-        >>> ns = Namespace(u'test/reviews')
-        >>> ns.create(u'A place for my reviews')
-        >>> ns.delete()
+        """Delete this namespace.
         """
         return self.api.delete()
 
     def tag(self, name):
-        """Get a child Tag
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.tag(u'about')
-        <fom.mapping.Tag object at 0x87c708c>
+        """Get a child tag of this namespace.
 
         :param name: The name of the child Tag to get.
         """
         return Tag(path_child(self.path, name))
 
     def namespace(self, name):
-        """Get a child Namespace
+        """Get a child namespace of this namespace
 
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.namespace(u'default')
-        <fom.mapping.Namespace object at 0x87c71ac>
-
-        :param name: The name of the child Namespace to get.
+        :param name: The name of the child namespace to get.
         """
         return Namespace(path_child(self.path, name))
 
@@ -165,11 +169,8 @@ class Namespace(SessionBound):
     def _get_description(self):
         """The description of this Namespace.
 
-        Setting the value attempts to set the description in the FluidDB.
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.description
-        u"FluidDB admin user's top-level namespace."
+        Setting and getting the value attempts to set/get the description in
+        FluidDB.
         """
         r = self.api.get(returnDescription=True)
         return r.value[u'description']
@@ -179,15 +180,13 @@ class Namespace(SessionBound):
         """
         return self.api.put(description)
 
-    description = property(_get_description, _set_description)
+    # python2.5
+    description = property(_get_description, _set_description,
+                           doc=_get_description.__doc__)
 
     @property
     def tag_names(self):
         """The names of the Tags in this Namespace.
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.tag_names
-        [u'about']
         """
         r = self.api.get(returnTags=True)
         return r.value[u'tagNames']
@@ -195,33 +194,18 @@ class Namespace(SessionBound):
     @property
     def tag_paths(self):
         """The full paths of the Tags in this Namespace
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.tag_paths
-        [u'fluiddb/about']
         """
-        return [
-            path_child(self.path, child) for child in self.tag_names
-        ]
+        return [path_child(self.path, child) for child in self.tag_names]
 
     @property
     def tags(self):
         """The Tags in this Namespace
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.tags
         """
-        return [
-            Tag(path) for path in self.tag_paths
-        ]
+        return [Tag(path) for path in self.tag_paths]
 
     @property
     def namespace_names(self):
         """The names of the child Namespaces in this Namespace
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.namespace_names
-        [u'default', u'tags', u'tag-values', u'namespaces', u'users']
         """
         r = self.api.get(returnNamespaces=True)
         return r.value[u'namespaceNames']
@@ -229,26 +213,15 @@ class Namespace(SessionBound):
     @property
     def namespace_paths(self):
         """The full paths of the child Namespaces in this Namespace
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.namespace_paths
-        [u'fluiddb/default', u'fluiddb/tags', u'fluiddb/tag-values' ...]
         """
-        return [
-            path_child(self.path, child) for child in self.namespace_names
-        ]
+        return [path_child(self.path, child) for child in self.namespace_names]
 
     @property
     def namespaces(self):
         """The child Namespaces in this Namespace
-
-        >>> ns = Namespace(u'fluiddb')
-        >>> ns.namespaces
-        [<fom.mapping.Namespace object at 0x877ea4c>, ...]
         """
-        return [
-            Namespace(path) for path in self.namespace_paths
-        ]
+        return [Namespace(path) for path in self.namespace_paths]
+
 
 class Tag(SessionBound):
     """A FluidDB Tag"""
@@ -256,7 +229,8 @@ class Tag(SessionBound):
     def __init__(self, path, fluid=None):
         super(Tag, self).__init__(path, fluid)
         self.permissions = Permissions(self.fluid.permissions.tags[self.path])
-        self.value_permissions = Permissions(self.fluid.permissions.tag_values[self.path])
+        self.value_permissions = Permissions(
+            self.fluid.permissions.tag_values[self.path])
 
     @property
     def api(self):
@@ -296,68 +270,193 @@ class Tag(SessionBound):
         """
         self.api.delete()
 
+
 class readonly_tag_value(object):
     """Descriptor to provide a tag value lookup on an object to simulate a
     simple attribute.
+
+    :param tagpath: The path of the tag to store this value in and read it
+                    from
     """
 
-    def __init__(self, tagpath):
+    def __init__(self, tagpath, cached=True):
         self.tagpath = tagpath
+        self.cached = cached
 
     def __get__(self, instance, owner):
-        return instance.get(self.tagpath)[0]
+        if self.cached:
+            value = instance.get_cached(self.tagpath)
+        else:
+            value = None
+        if value is None:
+            value = instance.get(self.tagpath)[0]
+        return value
 
 
 class tag_value(readonly_tag_value):
     """Descriptor to provide a tag value lookup on an object to simulate a
     simple attribute. With write support.
+
+    :param tagpath: The path of the tag to store this value in and read it
+                    from
+
+    :param content_type: The default MIME-type to be used when PUTting the
+                         value
+
+    :param cached: Flag to indicate if the value should be cached locally
+
+    :param lazy_save: Flag to indicate if the field's value should be updated
+                      when the instance.save() method is called
     """
 
-    def __init__(self, tagpath, defaultType=None):
+    def __init__(self, tagpath, content_type=None, cached=True,
+        lazy_save=True):
         """
-        defaultType argument sets the default mime-type to be used when
+        content_type argument sets the default mime-type to be used when
         PUTting the value
         """
-        super(tag_value, self).__init__(tagpath)
-        self.defaultType = defaultType
+        super(tag_value, self).__init__(tagpath, cached)
+        self.content_type = content_type
+        self.lazy_save = lazy_save
 
     def __set__(self, instance, value):
-        return instance.set(self.tagpath, value, self.defaultType)
-
+        if self.lazy_save:
+            instance.set_lazy_tag_value(self, value)
+        else:
+            return instance.set(self.tagpath, value, self.content_type)
 
 
 class Object(SessionBound):
     """An object
     """
+
     about = readonly_tag_value(u'fluiddb/about')
 
-    def __init__(self, uid=None, about=None, fluid=None):
+    def __init__(self, uid=None, about=None, fluid=None, initial={}):
+        # the object's UUID
         self.uid = uid
         self.fluid = fluid or Fluid.bound
+        # Create the _path_map dict for the instantiated class
+        self._path_map = {}
+        for attribute, tag in _get_tag_values(self.__class__):
+            self._path_map[tag.tagpath] = attribute
+        # check about isn't in the initial values
+        if 'fluiddb/about' in initial:
+            if about is None:
+                about = initial['fluiddb/about']['value']
+                del initial['fluiddb/about']
+        # react appropriately to an about value
         if about is not None:
             self.create(about)
+        # cache for tag values
+        self._cache = {}
+        # a list of fields whose value has been updated but not saved
+        self._dirty_fields = set()
+        # if there are some initial values then set them for the appropriate
+        # fields in this object (using the _path_map created above)
+        # keys = tag paths, e.g. "fluiddb/about"
+        # values = initial values, e.g. any FluidDB primitive type value
+        # see: http://api.fluidinfo.com/html/api.html#values_GET
+        for tag_path, value in initial.iteritems():
+            if 'value' in value:
+                setattr(self, self._path_map[tag_path], value['value'])
 
     def create(self, about=None):
         """Create a new object.
         """
         r = self.fluid.objects.post(about)
         self.uid = r.value[u'id']
+        self.about = about
 
-    def get(self, tag):
+    def get(self, tagpath):
         """Get the value of a tag.
         """
-        tagpath = tag
         r = self.api[tagpath].get()
+        self._cache[tagpath] = r.value
         return r.value, r.content_type
 
+    def get_cached(self, tagpath):
+        """Get the cached value of a tag.
+        """
+        return self._cache.get(tagpath)
+
+    def refresh(self, *tagpaths):
+        """
+        Clears the local cache
+        """
+        if tagpaths:
+            for tagpath in tagpaths:
+                del self._cache[tagpath]
+        else:
+            self._cache.clear()
+
     def set(self, tagpath, value, valueType=None):
         """Set the value of a tag.
         """
-        self.api[tagpath].put(value, valueType)
+        self._cache[tagpath] = value
+        # check if updating a tag handled by one of the tag_value attributes
+        if tagpath in self._path_map:
+            # get the tag_value instance
+            tv = self.__class__.__dict__[self._path_map[tagpath]]
+            if tv.lazy_save:
+                # update on save()
+                self.set_lazy_tag_value(tv, value)
+            else:
+                # update right now
+                self.api[tagpath].put(value, valueType)
+        else:
+            # the tag isn't associated with a tag_value on this object
+            self.api[tagpath].put(value, valueType)
+
+    def set_lazy_tag_value(self, tag_value, value):
+        """Sets the value of the given tag_value instance to be pushed to
+        FluidDB only when save() is called
+        """
+        # validate that the value is a primitive type (the only sort allowed
+        # by the /values api)
+        pt = type(value)
+        if pt in SERIALIZABLE_TYPES:
+            if pt in ITERABLE_TYPES:
+                if not all(isinstance(x, basestring) for x in value):
+                    # it was a list like structure that contained more than
+                    # just strings
+                    raise ValueError('Cannot lazy-save a non-primitive value.')
+        else:
+            raise ValueError('Cannot lazy-save a non-primitive value.')
+        # store away for when the save() method is called
+        self._dirty_fields.add(tag_value)
+        self._cache[tag_value.tagpath] = value
 
     def delete(self, tagpath):
+        """Removes a tag from the object
+        """
         self.api[tagpath].delete()
 
+    def save(self):
+        """Saves those fields that have been updated
+        """
+        if not self.about:
+            raise ValueError(
+                "Cannot save for an object without an about value")
+        if self._dirty_fields:
+            # values is the dict that will become the PUT payload to /values
+            values = {}
+            for item in self._dirty_fields:
+                tagpath = item.tagpath
+                # This check is done so the tag_relational capabilities work
+                # properly (i.e. the __get__ will return an instance of an
+                # object whose UUID should be referenced)
+                val = item.__get__(self, self.__class__)
+                if isinstance(item, tag_relation):
+                    val = val.uid
+                values[tagpath] = {'value': val}
+            # use the unique about value to identify this object in FluidDB
+            query = 'fluiddb/about = "%s"' % self.about
+            # update the values using the /values api
+            self.fluid.values.put(query, values)
+            # none of the fields are now dirty
+            self._dirty_fields.clear()
+
     @property
     def api(self):
         """The api ObjectApi for this instance.
@@ -400,7 +499,11 @@ class Object(SessionBound):
         return [class_type(uid) for uid in objects.value['ids']]
 
     def __repr__(self):
-        return '<%s %s>' % (self.__class__.__name__, self.about)
+        return '<%s %s>' % (self.__class__.__name__, self.uid)
+
+    def __eq__(self, other):
+        return self.uid and (other.uid == self.uid)
+
 
 class tag_relation(tag_value):
     """Descriptor to provide a relation lookup.
@@ -408,8 +511,8 @@ class tag_relation(tag_value):
     An id is actually stored in the database.
     """
 
-    def __init__(self, tag, object_type=Object):
-        tag_value.__init__(self, tag)
+    def __init__(self, tag, object_type=Object, cached=True, lazy_save=True):
+        tag_value.__init__(self, tag, cached=cached, lazy_save=lazy_save)
         self.object_type = object_type
 
     def __get__(self, instance, owner):
@@ -420,19 +523,52 @@ class tag_relation(tag_value):
         return tag_value.__set__(self, instance, value.uid)
 
 
+class tag_relations(tag_value):
+    """Descriptor to provide a list of relations.
+
+    A list of ids are stored in the database
+    """
+
+    def __init__(self, tagpath, object_type=Object, cached=True,
+            lazy_save=True):
+        tag_value.__init__(self, tagpath, cached=cached, lazy_save=lazy_save)
+        self.object_type = object_type
+
+    def __get__(self, instance, owner):
+        uids = tag_value.__get__(self, instance, owner)
+        return [self.object_type(uid) for uid in uids]
+
+    def __set__(self, instance, value):
+        uids = [obj.uid for obj in value]
+        return tag_value.__set__(self, instance, uids)
+
+
 class CollectionManager(object):
 
-    def __init__(self, instance, base_tagpath, object_type=Object):
+    base_nspath = 'test/fom/Collections'
+
+    def __init__(self, instance, tagpath, object_type, foreign_tagpath):
         self.instance = instance
-        self.base_tagpath = base_tagpath
+        self.tagpath = tagpath
         self.object_type = object_type
+        self.foreign_tagpath = foreign_tagpath
         self.target_tagpath = self._get_tagpath()
 
     def add(self, other):
         other.set(self.target_tagpath, self.instance.uid)
+        if self.foreign_tagpath is not None:
+            # don't set foreign on the reverse side
+            manager = CollectionManager(other, self.foreign_tagpath,
+                                        self.object_type, None)
+            manager.add(self.instance)
 
     def remove(self, other):
         other.delete(self.target_tagpath)
+        if self.foreign_tagpath is not None:
+            # don't set foreign on the reverse side
+            manager = CollectionManager(other, self.foreign_tagpath,
+                                        self.object_type, None)
+            manager.remove(self.instance)
 
     def __iter__(self):
         for item_id in self._fetch():
@@ -449,22 +585,23 @@ class CollectionManager(object):
 
     def _get_tagpath(self):
         try:
-            return self.instance.get(self.base_tagpath)[0]
+            return self.instance.get(self.tagpath)[0]
         except Fluid404Error:
             # manager is not yet created
             return self._create_manager()
 
     def _create_manager(self):
         uid = self._generate_uid()
-        tag_path = '/'.join([self.base_tagpath, uid])
-        self.instance.set(self.base_tagpath, tag_path)
-        Namespace(self.base_tagpath).create_tag(uid, u'Manager tag', True)
-        return tag_path
+        tagpath = '/'.join([self.base_nspath, uid])
+        self.instance.set(self.tagpath, tagpath)
+        Namespace(self.base_nspath).create_tag(uid, u'Manager tag', True)
+        return tagpath
 
     def _generate_uid(self):
-        url = '%s/%s' % (self.instance.fluid.db._get_url(self.instance.api.url),
-                         self.base_tagpath)
-        uid = unicode(uuid.uuid5(uuid.NAMESPACE_URL, str(url))).replace('-', '')
+        url = '%s/%s' % (self.instance.fluid.db._get_url(
+                            self.instance.api.url), self.tagpath)
+        uid = unicode(
+            uuid.uuid5(uuid.NAMESPACE_URL, str(url))).replace('-', '')
         return uid
 
     def _fetch(self):
@@ -472,18 +609,25 @@ class CollectionManager(object):
         for obj in resp.value[u'ids']:
             yield obj
 
-class tag_manager(tag_value):
+    def __str__(self):
+        return str(list(self))
 
-    def __init__(self, tagpath, map_type=Object, manager_type=CollectionManager):
-        tag_value.__init__(self, tagpath)
+
+class tag_collection(tag_value):
+
+    def __init__(self, tagpath, map_type=Object,
+                 manager_type=CollectionManager, foreign_tagpath=None,
+                 lazy_save=True):
+        tag_value.__init__(self, tagpath, lazy_save=True)
         self.map_type = map_type
         self.manager_type = manager_type
+        self.foreign_tagpath = foreign_tagpath
 
     def __get__(self, instance, owner):
         if instance.uid is None:
             raise ValueError(u'This object has not been created.')
-        return self.manager_type(instance, self.tagpath, self.map_type)
-
+        return self.manager_type(instance, self.tagpath, self.map_type,
+                                 self.foreign_tagpath)
 
 
 def path_child(path, child):
